"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

from wiliot_core.packet_data.config_files.parse_packet import *


class ConvertCloudPacket(object):
    def __init__(self, enriched_packet_row, original_packet=None, first_timestamp=0):
        """
        :param enriched_packet_row is one row from the enrich packet table from DB
        :type enriched_packet_row pd.Series
        """
        self.row = enriched_packet_row
        self.first_time = first_timestamp
        self.packet_length_type = 'UNKNOWN'
        if not ('rawPacket' in self.row.keys() and ('packetVersion' in self.row.keys() or 'flowVersion' in self.row.keys())):
            raise KeyError('cannot convert cloud packet without the following fields: '
                            '[rawPacket and packetVersion] or [rawPacket and flowVersion]')

        flow_ver = self.get_flow_ver()
        self.gw_packet = self.get_gw_data()
        base_fields = self.get_fields_before_payload(original_packet=original_packet, flow_ver=flow_ver)
        packet_out = self.combined_packet_parts(base_fields)
        packet_fixed = self.get_packet_data(packet_out)
        self.base_fields = base_fields
        self.reconstruct_packet = packet_fixed

    def get_reconstruct_packet(self):
        return self.reconstruct_packet

    def get_flow_ver(self):
        flow_ver_hex = '****'
        if 'flowVersion' in self.row:
            flow_ver_hex = str(self.row['flowVersion'])
            flow_ver_hex = flow_ver_hex.upper().replace('.', '').replace('X', '').zfill(FLOW_VER_SIZE)
            if len(flow_ver_hex) != FLOW_VER_SIZE:
                raise ValueError(f'ConvertCloudPacket: get_flow_ver: invlid flow version: {self.row["flowVersion"]}')
        return flow_ver_hex

    def get_fields_before_payload(self, original_packet, flow_ver):
        if original_packet is None:
            adva = flow_ver[:int(FLOW_VER_SIZE/2)] + \
                   ''.join(['*'] * (find_len(packet_data_info, 'adv_address') - FLOW_VER_SIZE)) + \
                   flow_ver[-int(FLOW_VER_SIZE/2):]
            en = ''.join(['*'] * find_len(packet_data_info, 'en'))
            b_type = ''.join(['*'] * find_len(packet_data_info, 'type'))
            e_header = ''.join(['*']) * find_len(packet_data_info, 'extended_header')
            adi = ''.join(['*']) * find_len(packet_data_info, 'adi')
            gw_packet = self.gw_packet
        else:
            adva = original_packet.packet_data['adv_address']
            en = original_packet.packet_data['en']
            b_type = original_packet.packet_data['type']
            e_header = original_packet.packet_data['extended_header'] \
                if 'extended_header' in original_packet.packet_data.keys() else \
                ''.join(['*']) * find_len(packet_data_info, 'extended_header')
            adi = original_packet.packet_data['adi'] \
                if 'adi' in original_packet.packet_data.keys() else \
                ''.join(['*']) * find_len(packet_data_info, 'adi')
            gw_packet = np.take(original_packet.gw_data['gw_packet'], 0)

        return {'adva': adva, 'en': en, 'b_type': b_type, 'e_header': e_header, 'adi': adi, 'gw_packet': gw_packet}

    def get_gw_data(self):
        gw_fields = {}
        gw_data = ''
        if 'rssi' in self.row:
            gw_fields['rssi'] = abs(int(self.row['rssi']))
        if 'timestamp' in self.row:
            gw_fields['gw_clock'] = int(self.row['timestamp']) - self.first_time

        if gw_fields:
            gw_bit_data = ''
            gw_fields['crc_valid'] = 1
            total_bits = 0
            for field, bit_len in gw_bit_structure.items():
                if field in gw_fields.keys():
                    bin_field = bin(int(gw_fields[field]))[2:]
                else:
                    bin_field = '0'
                bin_field = bin_field.zfill(bit_len)
                bin_field = bin_field[-bit_len:]
                gw_bit_data += str(bin_field)
                total_bits += bit_len
            gw_data = hex(int(gw_bit_data, 2))[2:].upper().zfill(int(total_bits / N_BITS_PER_CHAR))
        return gw_data
    
    @staticmethod
    def get_cloud_packet_length_type(packet_in):
        packet_len = len(packet_in)
        if packet_len == BLE_PAYLOAD_LENGTH:
            return 'BLE4_BLE5_PACKET'
        if packet_len == BLE_PAYLOAD_LENGTH - EN_TYPE_LEN:
            return 'BLE4_BLE5_PACKET'
        if packet_len == BLE_DOUBLE_PAYLOAD_LENGTH:
            return 'BLE5_DOUBLE_PACKET'
        estimated_bytes = int(packet_in[:2], 16)
        if estimated_bytes == len(packet_in[2:]) / 2:
            n_bytes = int(packet_in[:2],16) - int(TAG_EXTENDED_PAYLOAD_BYTES, 16)
            return f'BLE5_DOUBLE_PACKET_WITH_SIDE_INFO_{n_bytes}'  # including side info
        return 'UNKNOWN'

    def combined_packet_parts(self, all_fields):
        cloud_packet = str(self.row['rawPacket']).upper()
        if not is_cloud_packet(cloud_packet):
            raise Exception(f'cloud packet must starts with WILIOT_EN_TYPE: {WILIOT_EN_TYPE} or WILIOT_DATA_UID: {WILIOT_DATA_UID}')
        
        if any(cloud_packet.startswith(x) for x in WILIOT_DATA_UID):
            cloud_packet = all_fields['en'] + all_fields['b_type'] + cloud_packet

        self.packet_length_type = self.get_cloud_packet_length_type(cloud_packet)
        if self.packet_length_type == 'BLE4_BLE5_PACKET':
            return all_fields['adva'] + cloud_packet + all_fields['gw_packet']
        
        elif 'BLE5_DOUBLE_PACKET' in self.packet_length_type:
            header = [k for k, v in packet_length_types.items() if v['name'] == 'BLE5-DBL-EXT']
            phy_header = header[0]
            if 'SIDE_INFO' in self.packet_length_type:
                phy_header = header[0][:2] + hex(int(header[0][2:], 16) + int(self.packet_length_type.split('_')[-1]))[2:].upper().zfill(2)
            return phy_header + all_fields['e_header'] + all_fields['adva'] + all_fields['adi'] + cloud_packet + all_fields['gw_packet']
        else:
            raise Exception(f'unsupported packet length: packet length: {len(cloud_packet)} and supported lengths are: {[BLE_PAYLOAD_LENGTH, BLE_DOUBLE_PAYLOAD_LENGTH]}')

    def get_packet_data(self, packet_out):
        new_packet = packet_out
        if 'decryptedData' in self.row.keys() and 'tagId' in self.row.keys():
            # decrypted data
            dec_tag_id = reverse_data(str(self.row['tagId']))
            dec_payload = str(self.row['decryptedData']).upper()

            type_str = 'BLE5-DBL-EXT' if 'BLE5_DOUBLE_PACKET' in self.packet_length_type else 'LEGACY'
            
            # update fields
            fields_to_update = {'enc_uid': dec_tag_id,
                                'enc_payload': dec_payload,
                                }
            for field, field_value in fields_to_update.items():
                start, length = packet_data_info[field]['index_info'][type_str]
                new_packet = new_packet[:start] + field_value.upper() + new_packet[start + length:]
            
            # zero mic to indicated decrypted packet:
            mic_start, mic_length = packet_data_info['mic']['index_info'][type_str]
            if type_str == 'LEGACY':
                mic_start += 6 # only last 3 bytes used for decryption
                mic_length = 6
            dec_mic = '0' * mic_length
            new_packet = new_packet[:mic_start] + dec_mic + new_packet[mic_start + mic_length:]
        return new_packet



"""
Copyright (c) 2016- 2025, Wiliot Ltd. All rights reserved.

Redistribution and use of the Software in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

  1. Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

  2. Redistributions in binary form, except as used in conjunction with
  Wiliot's Pixel in a product or a Software update for such product, must reproduce
  the above copyright notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the distribution.

  3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
  may be used to endorse or promote products or services derived from this Software,
  without specific prior written permission.

  4. This Software, with or without modification, must only be used in conjunction
  with Wiliot's Pixel or with Wiliot's cloud service.

  5. If any Software is provided in binary form under this license, you must not
  do any of the following:
  (a) modify, adapt, translate, or create a derivative work of the Software; or
  (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
  discover the source code or non-literal aspects (such as the underlying structure,
  sequence, organization, ideas, or algorithms) of the Software.

  6. If you create a derivative work and/or improvement of any Software, you hereby
  irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
  royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
  right and license to reproduce, use, make, have made, import, distribute, sell,
  offer for sale, create derivative works of, modify, translate, publicly perform
  and display, and otherwise commercially exploit such derivative works and improvements
  (as applicable) in conjunction with Wiliot's products and services.

  7. You represent and warrant that you are not a resident of (and will not use the
  Software in) a country that the U.S. government has embargoed for use of the Software,
  nor are you named on the U.S. Treasury Department’s list of Specially Designated
  Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
  You must not transfer, export, re-export, import, re-import or divert the Software
  in violation of any export or re-export control laws and regulations (such as the
  United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
  and use restrictions, all as then in effect

THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
(SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
(A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
(B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
(C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

import numpy as np
import pandas as pd

from wiliot_core.packet_data.config_files.utils_map import get_nonce_minus_one


class StatisticsCalc(object):
    def __init__(self, packet_df=None, sprinkler_df=None, is_enriched=False):
        """
        :type packet_df pd.DataFrame or None
        :type sprinkler_df pd.DataFrame or None
        """
        self.packet_df = packet_df if packet_df is not None else pd.DataFrame()
        self.sprinkler_df = sprinkler_df if sprinkler_df is not None else pd.DataFrame()
        self.is_enriched = is_enriched

        self.n_bad_packets = 0
        if 'crc_valid' in self.packet_df:
            self.packet_df = self.packet_df[self.packet_df['crc_valid'] == 1]
            self.n_bad_packets = (packet_df['crc_valid'] == 0).sum()
        if 'crc_valid' in self.sprinkler_df:
            self.sprinkler_df = self.sprinkler_df[self.sprinkler_df['crc_valid'] == 1]

        self.packet_df.reset_index(inplace=True, drop=True)
        self.sprinkler_df.reset_index(inplace=True, drop=True)

        self.n_packets = self.packet_df.shape[0]
        self.n_cycles = self._get_n_cycles()

    def _get_n_cycles(self):
        if self.sprinkler_df.shape[0] <= 1:
            return self.sprinkler_df.shape[0]
        flow_ver_dict = self._get_unique_list(param='flow_ver')
        flow_ver = flow_ver_dict['flow_ver'].lower()
        if flow_ver == '':
            return 0
        flow_ver = flow_ver[2:] if flow_ver.startswith('0x') else flow_ver.replace('.', '')
        if flow_ver < '60a':  # gen2
            return self.sprinkler_df.shape[0]
        # gen3:
        if 'decrypted_packet_type' not in self.sprinkler_df:
            return 0
        all_nonce = pd.DataFrame()
        for packet_type_num, df_type in self.sprinkler_df.groupby('decrypted_packet_type'):
            if packet_type_num == 3:
                continue
            if packet_type_num == 1:
                new_nonce = df_type['nonce'].apply(lambda x: get_nonce_minus_one(x))
            else:
                new_nonce = df_type['nonce']
            all_nonce = pd.concat([all_nonce, new_nonce], ignore_index=True)
        if all_nonce.shape[0] <= 1:
            return all_nonce.shape[0]
        return int(all_nonce.nunique().values[0])

    @staticmethod
    def _get_stat(col_in, nan_val=None, added_val=None, added_n=None):
        """
        :param col_in the array for the statistics
        :type col_in pd.Series or np.ndarray or list
        :param nan_val value of the nan statistics, the default is float('nan')
        :type None or float or int
        :param added_val list of values to add to the statistics, instead of add it to the col_in to improve performance
        :type added_val list
        :param added_n corresponding to added_val, list of the amount of each value from add_val that is needed to be
                       add to the statistics, instead of add it to the col_in to improve performance
        :type added_n list
        """
        if nan_val is None:
            nan_val = float('nan')

        if col_in is None or all(pd.isnull(col_in)):
            return {'mean': nan_val,
                    'std': nan_val,
                    'min': nan_val,
                    'max': nan_val}
        if added_val is None or added_val == [] or added_n is None or added_n == []:
            return {'mean': np.nanmean(col_in),
                    'std': np.nanstd(col_in),
                    'min': np.nanmin(col_in),
                    'max': np.nanmax(col_in)}
        n = sum(added_n + [np.count_nonzero(~np.isnan(col_in))])
        total_val = sum([a_v * a_n for a_v, a_n in zip(added_n, added_val)])
        return {'mean':  (np.nansum(col_in) + total_val) / n,
                'std': -1.0,
                'min': min([np.nanmin(col_in)] + added_val),
                'max': max([np.nanmax(col_in)] + added_val)}

    def _get_unique_list(self, param, param_name=None):
        if self.sprinkler_df.empty or param not in self.sprinkler_df:
            return {param_name or param: ''}
        if param_name is None:
            param_name = param
        return {param_name: ','.join(self.sprinkler_df[param].unique().astype(str))}

    def _get_last_if_exists(self, name):
        if name in self.sprinkler_df:
            data = self.sprinkler_df[name].dropna()
            if len(data):
                return data.iloc[-1]
        return None

    def get_counting(self):
        return {'num_packets': self.n_packets,
                'num_cycles': self.n_cycles,
                'num_bad_crc_packets': self.n_bad_packets
                }

    def get_sprinkler_counter_stat(self):
        col_in = None
        if 'num_packets_per_sprinkler' in self.sprinkler_df:
            col_in = self.sprinkler_df['num_packets_per_sprinkler']
        stat = self._get_stat(col_in)
        return {f'sprinkler_counter_{k}': v for k, v in stat.items()}

    def get_tbp_stat(self):
        tbp_for_calc = None
        if 'tbp' in self.sprinkler_df:
            tbp_for_calc = self.sprinkler_df['tbp'].replace(to_replace=-1, value=np.nan, inplace=False)
        stat = self._get_stat(tbp_for_calc, -1)
        stat_out = {f'tbp_{k}': v for k, v in stat.items()}
        stat_out['tbp_num_vals'] = sum(~(pd.isnull(tbp_for_calc))) if tbp_for_calc is not None else 0
        return stat_out

    def get_per_stat(self):
        stat_param = ['mean', 'std']
        col_in = None
        if 'per' in self.sprinkler_df:
            col_in = self.sprinkler_df['per']
        stat = self._get_stat(col_in)
        stat_out = {f'per_{k}': v for k, v in stat.items() if k in stat_param}

        return stat_out

    def get_rssi_stat(self):
        col_in = None
        if 'rssi' in self.packet_df:
            col_in = self.packet_df['rssi']
        stat = self._get_stat(col_in)
        return {f'rssi_{k}': v for k, v in stat.items()}

    def get_time_stat(self):
        col_in = None
        if 'time_from_start' in self.packet_df:
            col_in = self.packet_df['time_from_start']

        stat = self._get_stat(col_in)
        stat_out = {'ttfp': stat['min'],
                    'end_time': stat['max'],
                    'duration': stat['max'] - stat['min']}
        if stat_out.get('duration', 0) > 0:
            stat_out['rx_rate'] = self.n_packets / stat_out['end_time']
            stat_out['rx_rate_normalized'] = self.n_packets / stat_out['duration']
        else:
            stat_out['rx_rate'] = float('nan')
            stat_out['rx_rate_normalized'] = float('nan')
        return stat_out

    def get_flow_ver(self):
        return self._get_unique_list(param='flow_ver')

    def get_asset_id(self):
        if 'asset_id' in self.sprinkler_df:
            return self._get_unique_list(param='asset_id', param_name='asset_id')
        if 'assetId' in self.sprinkler_df:
            return self._get_unique_list(param='assetId', param_name='asset_id')
        return {'asset_id': ''}
    
    def get_external_id(self):
        if 'external_id' in self.sprinkler_df:
            return self._get_unique_list(param='external_id', param_name='external_id')
        if 'externalId' in self.sprinkler_df:
            return self._get_unique_list(param='externalId', param_name='external_id')
        return {'external_id': ''}
    
    # ############ Function only for Enriched (after parse/decryption API)  Packets ###############
    def get_min_tx_stat(self):
        if not self.is_enriched:
            return {}
        col_in = None
        if 'min_tx' in self.sprinkler_df:
            col_in = self.sprinkler_df['min_tx']
        stat_out = {}
        val = self._get_last_if_exists('min_tx')
        val = val if val is not None else float('nan')
        stat_out['min_tx_last'] = val
        stat = self._get_stat(col_in)
        stat_out['min_tx_mean'] = stat['mean']
        return stat_out

    def get_lo_max_freq_stat(self):
        if not self.is_enriched:
            return {}
        col_in = None
        if 'lo_max_freq_signed' in self.sprinkler_df and not all(self.sprinkler_df['lo_max_freq_signed'].isnull()):
            col_in = self.sprinkler_df['lo_max_freq_signed']
        elif 'lo_freq_max' in self.sprinkler_df:
            col_in = self.sprinkler_df['lo_freq_max']
        elif 'lo_max_freq' in self.sprinkler_df:
            col_in = self.sprinkler_df['lo_max_freq']
        if col_in is not None:
            col_in = col_in.apply(lambda x: x if x != 2402 else None)
        stat = self._get_stat(col_in)
        stat_out = {f'lo_max_freq_{k}': v for k, v in stat.items()}
        return stat_out